from aiohttp import web
from transmute_core.frameworks.aiohttp import (
    describe,
    add_swagger,
    add_route,
    route,
    APIException,
    Response,
)
from aiohttp.web import HTTPForbidden
from .utils import User


async def handle(request):
    text = "Hello, can you hear me?"
    return web.Response(body=text.encode("utf-8"))


@describe(paths="/multiply")
async def multiply(request, left: int, right: int) -> int:
    return left * right


@describe(paths="/id/{my_id}")
async def get_id(request, my_id: str) -> str:
    return "your id is: " + my_id


@describe(paths="/optional")
async def get_optional(request, include_foo: bool = False) -> bool:
    return include_foo


@describe(
    paths="/headers/",
    response_types={
        200: {
            "type": str,
            "description": "success",
            "headers": {
                "location": {"description": "url to the location", "type": str}
            },
        }
    },
)
async def header_response(request):
    return Response("foo", headers={"location": "boo"})


@describe(paths="/aiohttp_error")
async def error(request):
    raise HTTPForbidden(reason="unauthorized")


@describe(paths="/api_exception")
async def api_exception(request) -> User:
    raise APIException("nope")


@describe(
    paths="/body_and_header",
    methods="POST",
    body_parameters=["body"],
    header_parameters=["header"],
)
async def body_and_header(request, body: str, header: str) -> bool:
    return body == header


@describe(paths="/config")
async def config(request):
    return request.app["config"]


@describe(paths="/multiple_query_params")
async def multiple_query_params(tag: [str]) -> str:
    return ",".join(tag)


def create_app():
    app = web.Application()
    app["config"] = {"test": "foo"}
    app.router.add_route("GET", "/", handle)
    add_route(app, multiple_query_params)
    add_route(app, multiply)
    add_route(app, get_id)
    add_route(app, header_response)
    route(app, config)
    route(app, get_optional)
    route(app, body_and_header)
    route(app, error)
    route(app, api_exception)
    # this should be at the end, to ensure all routes are considered when
    # constructing the handler.
    add_swagger(app, "/swagger.json", "/swagger")
    return app



import tornado.web

METHODS = ["get", "head", "post", "delete", "patch", "put", "options"]


def url_spec(transmute_path, handler, *args, **kwargs):
    """
    convert the transmute_path
    to a tornado compatible regex,
    and return a tornado url object.
    """
    p = _to_tornado_pattern(transmute_path)
    for m in METHODS:
        method = getattr(handler, m)
        if hasattr(method, "transmute_func"):
            method.transmute_func.paths.add(transmute_path)
    return tornado.web.URLSpec(p, handler, *args, **kwargs)


def _to_tornado_pattern(transmute_path):
    """convert a transmute path to
    a tornado pattern.
    """
    return (
        transmute_path.replace("{", "(?P<")
        # .replace("}", ">[^\/]+)"))
        .replace("}", ">.*)")
    )



from transmute_core.frameworks.flask import (
    route,
    annotate,
    Response,
    APIException,
    add_swagger,
)
from schematics.models import Model
from schematics.types import StringType, IntType
from flask import Flask, Blueprint

app = Flask(__name__)


class Person(Model):
    name = StringType()
    number_of_cats = IntType()


CAT_PEOPLE = {"george": Person({"name": "george", "number_of_cats": 2})}


@route(app, paths="/cat_person")
@annotate({"name": str, "return": Person})
def get_cat_person(name):
    """find and return a cat person"""
    if name in CAT_PEOPLE:
        return CAT_PEOPLE[name]
    raise APIException("no cat person found", code=404)


# alternatively
@route(app, paths="/multiply")
@annotate({"left": int, "right": int, "return": int})
def multiply(left, right):
    """multiply two integers"""
    return left * right


# type hint equivalent in Py2
multiply.__annotations__ = {"left": int, "right": int, "return": int}


@route(app, paths="/exception")
def exception():
    raise APIException("api error!")


@route(
    app,
    paths="/complex/{path}",
    methods=["POST"],
    body_parameters=["body"],
    header_parameters=["header"],
)
@annotate({"body": str, "header": str, "path": str, "return": str})
def complex(body, header, path):
    return body + ":" + header + ":" + path


blueprint = Blueprint("blueprint", __name__, url_prefix="/blueprint")


@route(blueprint, paths="/foo")
@annotate({"return": bool})
def foo():
    return True


@route(blueprint, paths="/test/", methods=["post"], body_parameters="body")
@annotate({"return": bool, "body": Person})
def test(body):
    return True


@route(
    app,
    paths="/api/v1/header",
    response_types={
        200: {
            "type": str,
            "description": "success",
            "headers": {
                "location": {"description": "url to the location", "type": str}
            },
        },
    },
)
def header():
    return Response("foo", headers={"x-nothing": "value"})


app.register_blueprint(blueprint)

# finally, you can add a swagger json and a documentation page by:
add_swagger(app, "/swagger.json", "/api/")

if __name__ == "__main__":
    app.run(debug=True)



import pytest
import json


@pytest.mark.asyncio
async def test_content_type_in_response(cli):
    """the content type should be specified in the response."""
    resp = await cli.get("/optional")
    assert 200 == resp.status
    assert resp.headers["Content-Type"] == "application/json"


@pytest.mark.asyncio
async def test_headers(cli):
    """the content type should be specified in the response."""
    resp = await cli.get("/headers/")
    assert "boo" == resp.headers["location"]


@pytest.mark.asyncio
async def test_multiply(cli):
    resp = await cli.get("/multiply?left=5&right=10")
    assert 200 == resp.status
    text = await resp.text()
    assert json.loads(text) == 50


@pytest.mark.asyncio
async def test_multiply_bad_int(cli):
    resp = await cli.get("/multiply?left=foo&right=0x00")
    assert 400 == resp.status


@pytest.mark.asyncio
async def test_error(cli):
    """HTTPProcessingErrors should be honored."""
    resp = await cli.get("/aiohttp_error")
    assert 403 == resp.status
    text = await resp.text()
    assert "unauthorized" in json.loads(text)["result"]


@pytest.mark.asyncio
async def test_api_exception(cli):
    """APIException should be honored."""
    resp = await cli.get("/api_exception")
    assert 400 == resp.status
    text = await resp.text()
    assert "nope" in json.loads(text)["result"]


@pytest.mark.asyncio
async def test_optional(cli):
    resp = await cli.get("/optional")
    assert 200 == resp.status
    ret_value = await resp.json()
    assert ret_value is False


@pytest.mark.asyncio
async def test_optional_with_value(cli):
    resp = await cli.get("/optional?include_foo=true")
    assert 200 == resp.status
    ret_value = await resp.json()
    assert ret_value is True


@pytest.mark.asyncio
async def test_body_and_header(cli):
    resp = await cli.post(
        "/body_and_header",
        data=json.dumps({"body": "body"}),
        headers={"content-type": "application/json", "header": "header"},
    )
    assert 200 == resp.status
    ret_value = await resp.json()
    assert ret_value is False


@pytest.mark.asyncio
async def test_get_id(cli):
    resp = await cli.get("/id/10")
    assert 200 == resp.status
    text = await resp.text()
    assert json.loads(text) == "your id is: 10"


@pytest.mark.asyncio
async def test_config(cli):
    resp = await cli.get("/config")
    assert 200 == resp.status
    text = await resp.text()
    assert json.loads(text) == {"test": "foo"}


@pytest.mark.asyncio
async def test_swagger(cli):
    resp = await cli.get("/swagger.json")
    assert 200 == resp.status
    assert "application/json" == resp.headers["Content-Type"]
    text = await resp.text()
    assert json.loads(text)["paths"]["/multiply"]["get"]["responses"] == {
        "200": {"schema": {"type": "integer"}, "description": "success"},
        "400": {
            "schema": {
                "title": "FailureObject",
                "type": "object",
                "required": ["success", "result"],
                "properties": {
                    "result": {"type": "string"},
                    "success": {"type": "boolean"},
                },
            },
            "description": "invalid input received",
        },
    }


@pytest.mark.asyncio
async def test_swagger_page(cli):
    resp = await cli.get("/swagger")
    assert 200 == resp.status
    assert "text/html" == resp.headers["Content-Type"]



"""
tests for legacy url dispatcher.
"""
import pytest
from transmute_core.frameworks.aiohttp.url_dispatcher import TransmuteUrlDispatcher
from transmute_core.frameworks.aiohttp.swagger import get_swagger_spec
from transmute_core import describe
from unittest.mock import patch


async def describe_later() -> str:
    return "foo"


@describe(paths="/describe_now")
async def describe_now() -> str:
    return "foo"


@pytest.fixture
def url_dispatcher(app):
    dispatcher = TransmuteUrlDispatcher()
    dispatcher.post_init(app)
    return dispatcher


def test_urldispatcher_exception_invalid_argcount(url_dispatcher):
    with pytest.raises(ValueError):
        url_dispatcher.add_transmute_route(1, 2, 3, 4)


def test_urldispatcher_adds_route(app, url_dispatcher):
    with patch(
        "transmute_core.frameworks.aiohttp.url_dispatcher.add_route"
    ) as add_route:
        url_dispatcher.add_transmute_route("GET", "/describe_later", describe_later)
        assert "GET" in describe_later.transmute.methods
        assert "/describe_later" in describe_later.transmute.paths
        add_route.assert_called_with(
            app, describe_later, context=url_dispatcher._transmute_context
        )


def test_urldispatcher_adds_route_one_arg(app, url_dispatcher):
    with patch(
        "transmute_core.frameworks.aiohttp.url_dispatcher.add_route"
    ) as add_route:
        url_dispatcher.add_transmute_route(describe_now)
        add_route.assert_called_with(
            app, describe_now, context=url_dispatcher._transmute_context
        )



import pytest


@pytest.mark.asyncio
async def test_parsing_multiiple_query_params(cli):
    resp = await cli.get("/multiple_query_params?tag=foo&tag=bar")
    ret_value = await resp.json()
    assert 200 == resp.status
    assert ret_value == "foo,bar"


@pytest.mark.asyncio
async def test_parsing_multiple_query_params_single_tag(cli):
    resp = await cli.get("/multiple_query_params?tag=foo")
    ret_value = await resp.json()
    assert 200 == resp.status
    assert ret_value == "foo"



import sys
from functools import wraps
from flask import request, Response
from transmute_core import ParamExtractor, NoArgument


def create_routes_and_handler(transmute_func, context):
    @wraps(transmute_func.raw_func)
    def handler(*args, **kwargs):
        exc, result = None, None
        try:
            args, kwargs = _param_instance.extract_params(
                context,
                transmute_func,
                request.content_type,
            )
            result = transmute_func(*args, **kwargs)
        except Exception as e:
            exc = e
            exc.__traceback__ = sys.exc_info()[2]
        response = transmute_func.process_result(
            context, result, exc, request.content_type
        )
        return Response(
            response["body"],
            status=response["code"],
            mimetype=response["content-type"],
            headers=response["headers"],
        )

    return (_convert_paths_to_flask(transmute_func.paths), handler)


def _convert_paths_to_flask(transmute_paths):
    """flask has it's own route syntax, so we convert it."""
    paths = []
    for p in transmute_paths:
        paths.append(p.replace("{", "<").replace("}", ">"))
    return paths


class ParamExtractorFlask(ParamExtractor):
    def _get_framework_args(self):
        return {}

    @property
    def body(self):
        return request.get_data()

    @staticmethod
    def _query_argument(key, is_list):
        if key not in request.args:
            return NoArgument
        if is_list:
            return request.args.getlist(key)
        else:
            return request.args[key]

    @staticmethod
    def _header_argument(key):
        return request.headers.get(key, NoArgument)

    @staticmethod
    def _path_argument(key):
        return request.view_args.get(key, NoArgument)


_param_instance = ParamExtractorFlask()



import pytest
import sys


def pytest_ignore_collect(*args, **kwargs):
    if sys.version_info < (3, 5):
        return True
    return False


@pytest.fixture
def app(loop):
    from .example import create_app

    return create_app()


@pytest.fixture
def cli(app, loop, aiohttp_client):
    return loop.run_until_complete(aiohttp_client(app))



from transmute_core import *
from .url_dispatcher import TransmuteUrlDispatcher
from .swagger import add_swagger_api_route, create_swagger_json_handler, add_swagger
from .route import add_route

route = add_route



import warnings
from aiohttp.web import UrlDispatcher
from collections import OrderedDict
from transmute_core import default_context, describe
from transmute_core.function import TransmuteFunction
from transmute_core.swagger import SwaggerSpec
from .route import add_route


class TransmuteUrlDispatcher(UrlDispatcher):
    """
    .. warning:: The TransmuteUrlDispatcher is deprecated.
                 please use aiohttp_transmute.add_route instead.

    A UrlDispatcher which instruments the add_route function to
    collect swagger spec data from transmuted functions.
    """

    def __init__(self, *args, context=default_context, **kwargs):
        warnings.warn(
            (
                "TransmuteUrlDispatch has been deprecated. "
                " And will be removed in a future version. "
                " Please use aiohttp_transmute.add_route instead. "
            ),
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__()
        self._transmute_context = context

    def post_init(self, app):
        """
        this must be called after dispatcher creation, to ensure proper operability.
        """
        self._app = app

    def add_transmute_route(self, *args):
        """
        two formats are accepted, for transmute routes. One allows
        for a more traditional aiohttp syntax, while the other
        allows for a flask-like variant.

        .. code-block:: python

            # if the path and method are not added in describe.
            add_transmute_route("GET", "/route", fn)

            # if the path and method are already added in describe
            add_transmute_route(fn)
        """
        if len(args) == 1:
            fn = args[0]
        elif len(args) == 3:
            methods, paths, fn = args
            fn = describe(methods=methods, paths=paths)(fn)
        else:
            raise ValueError("expected one or three arguments for add_transmute_route!")

        add_route(self._app, fn, context=self._transmute_context)



from transmute_core import TransmuteFunction, default_context
from .handler import create_handler
from .swagger import get_swagger_spec


def add_route(app, fn, context=default_context):
    """
    a decorator that adds a transmute route to the application.
    """
    transmute_func = TransmuteFunction(fn, args_not_from_request=["request"])
    handler = create_handler(transmute_func, context=context)
    get_swagger_spec(app).add_func(transmute_func, context)

    for p in transmute_func.paths:
        aiohttp_path = _convert_to_aiohttp_path(p)
        resource = app.router.add_resource(aiohttp_path)
        for method in transmute_func.methods:
            resource.add_route(method, handler)


def _convert_to_aiohttp_path(path):
    """convert a transmute path to one supported by aiohttp."""
    return path



import copy
from swagger_schema import Schema


class ResponseShape(object):
    """
    result shapes define the return format of the
    response.
    """

    @staticmethod
    def create_body(result_dict):
        """
        given the result dict from
        transmute_func, return back the
        response object.
        """
        raise NotImplementedError()

    @staticmethod
    def swagger(result_schema):
        """
        given the schema of the inner
        result object, return back the
        swagger schema representation.
        """
        raise NotImplementedError()


class ResponseShapeSimple(ResponseShape):
    """return back just the result object."""

    @staticmethod
    def create_body(result_dict):
        return result_dict["result"]

    @staticmethod
    def swagger(result_schema):
        return result_schema


class ResponseShapeComplex(ResponseShape):
    """
    return back an object with the result nested,
    providing a little more context on the result:

    * status code
    * success
    * result
    """

    @staticmethod
    def create_body(result_dict):
        return result_dict

    @staticmethod
    def swagger(result_schema):
        return Schema(
            {
                "title": "SuccessObject",
                "type": "object",
                "properties": {
                    "result": result_schema,
                    "success": {"type": "boolean"},
                    "code": {"type": "number"},
                },
                "required": ["success", "result", "code"],
            }
        )



import inspect
from swagger_schema import Operation, Response, Schema, PathItem
from ..compat import getfullargspec
from ..context import default_context
from ..attributes import TransmuteAttributes, ResponseType
from .signature import FunctionSignature
from .parameters import get_parameters
from ..http_parameters import get_swagger_parameters
from ..exceptions import InvalidTransmuteDefinition
from ..handler import process_result


class TransmuteFunction(object):
    """
    TransmuteFunctions wrap a function and add metadata,
    allowing transmute frameworks to extract that information for
    their own use (such as web handler generation or automatic
    documentation)
    """

    params = ["error_exceptions"]

    def __init__(self, func, args_not_from_request=None):
        """
        :param list(str) args_not_from_request: a list of arguments that
            transmute_function should not retrieve from the request
            (e.g. from query parameters). This is useful for frameworks
            with pass in an argument to represent the request, or response object.
        """
        # arguments should be the arguments passed into
        # the function
        #
        # the return type should be specified. If nothing is
        # passed in, a NoneType is returned.
        # for a dynamic language like Python, it's not a huge deal:
        # transmute will still json serialize whatever object you pass it.
        # however, the return type is valuable for autodocumentation systems.
        attrs = getattr(func, "transmute", TransmuteAttributes())
        self._validate_attributes(attrs)
        # these are the paths that the transmute function
        # should respond to.
        self.paths = attrs.paths
        self.tags = attrs.tags
        self.success_code = attrs.success_code
        argspec = getfullargspec(func)
        self.signature = FunctionSignature.from_argspec(argspec)
        self.response_types = self._parse_response_types(argspec, attrs)
        self.parameters = get_parameters(
            self.signature, attrs, arguments_to_ignore=args_not_from_request
        )
        # description should be a description of the api
        # endpoint, for use in autodocumentation
        self.description = inspect.cleandoc(func.__doc__ or "")
        # summary should be a shortened version of description.
        self.summary = self._summary(self.description)
        # error_exceptions represents the exceptions
        # that should be caught and return an API exception
        self.error_exceptions = tuple(attrs.error_exceptions)
        # these are the http methods that should route to the given function.
        self.methods = attrs.methods
        self.raw_func = func
        # this is to make discovery easier.
        # TODO: make sure this doesn't mess up GC, as it's
        # a cyclic reference.
        if inspect.ismethod(func):
            func.__func__.transmute_func = self
        else:
            func.transmute_func = self

    def __call__(self, *args, **kwargs):
        return self.raw_func(*args, **kwargs)

    def get_response_by_code(self, code):
        """return the return type, by code"""
        return self.response_types.get(code, {}).type

    def get_swagger_path(self, context=default_context):
        operation = self.get_swagger_operation(context)
        path = PathItem()
        for m in self.methods:
            name = m.lower()
            if name == "get":
                # get is a reserved keyword in a schematics dict
                # https://github.com/schematics/schematics/pull/489
                name = "get_"
            setattr(path, name, operation)
        return path

    def get_swagger_operation(self, context=default_context):
        """
        get the swagger_schema operation representation.
        """
        consumes = produces = context.contenttype_serializers.keys()
        parameters = get_swagger_parameters(self.parameters, context)
        responses = {
            "400": Response(
                {
                    "description": "invalid input received",
                    "schema": Schema(
                        {
                            "title": "FailureObject",
                            "type": "object",
                            "properties": {
                                "success": {"type": "boolean"},
                                "result": {"type": "string"},
                            },
                            "required": ["success", "result"],
                        }
                    ),
                }
            )
        }

        for code, details in self.response_types.items():
            responses[str(code)] = details.swagger_definition(context)

        return Operation(
            {
                "summary": self.summary,
                "description": self.description,
                "consumes": consumes,
                "produces": produces,
                "parameters": parameters,
                "responses": responses,
                "operationId": self.raw_func.__name__,
                "tags": self.tags,
            }
        )

    def process_result(self, context, result_body, exc, content_type):
        """
        given an result body and an exception object,
        return the appropriate result object,
        or raise an exception.
        """
        return process_result(self, context, result_body, exc, content_type)

    @staticmethod
    def _validate_attributes(attrs):
        """
        validate a TransmuteAttributes contains valid configuration
        that can be converted into a TransmuteFunction.
        """
        if len(attrs.paths) == 0:
            raise InvalidTransmuteDefinition(
                "at least one path is required for a valid set of transmute attribute: {0}".format(
                    str(attrs)
                )
            )

    @staticmethod
    def _parse_response_types(argspec, attrs):
        """
        from the given parameters, return back the response type dictionaries.
        """
        return_type = argspec.annotations.get("return") or None
        type_description = attrs.parameter_descriptions.get("return", "")
        response_types = attrs.response_types.copy()
        if return_type or len(response_types) == 0:
            response_types[attrs.success_code] = ResponseType(
                type=return_type,
                type_description=type_description,
                description="success",
            )
        return response_types

    @staticmethod
    def _summary(description):
        for s in description.splitlines():
            if s != "":
                return s
        return ""



import sys
import tornado.gen
from functools import wraps
from transmute_core import (
    ParamExtractor,
    NoArgument,
    default_context,
    TransmuteFunction,
)


def convert_to_handler(context=default_context):
    def decorator(fn):
        func = tornado.gen.coroutine(fn)
        transmute_func = TransmuteFunction(fn)

        @tornado.gen.coroutine
        @wraps(transmute_func.raw_func)
        def wrapper(self, *passed_args, **kwargs):
            result, exc = None, None
            content_type = self.request.headers.get("Content-Type", None)
            param_extractor = ParamExtractorTornado(self, kwargs)
            try:
                args, kwargs = param_extractor.extract_params(
                    context, transmute_func, content_type
                )
                args = list(passed_args) + args
                result = yield func(self, *args, **kwargs)
            except Exception as e:
                exc = e
                exc.__traceback__ = sys.exc_info()[2]
            response = transmute_func.process_result(context, result, exc, content_type)
            self.set_header("Content-Type", response["content-type"])
            self.set_status(response["code"])
            self.finish(response["body"])

        wrapper.transmute_func = transmute_func
        return wrapper

    return decorator


class ParamExtractorTornado(ParamExtractor):
    def __init__(self, handler_self, path_kwargs):
        self._handler_self = handler_self
        self._request = handler_self.request
        self._path_kwargs = path_kwargs

    @property
    def body(self):
        return self._request.body

    def _query_argument(self, key, is_list):
        qa = self._request.query_arguments
        if key not in qa:
            return NoArgument
        if is_list:
            return [v.decode() for v in qa[key]]
        else:
            return qa[key][0].decode()

    def _header_argument(self, key):
        return self._request.headers.get(key, NoArgument)

    def _path_argument(self, key):
        return self._path_kwargs.get(key, NoArgument)

    def _get_framework_args(self):
        return {}



import tornado.testing
import tornado.web
from transmute_core.frameworks.tornado import RouteSet
from transmute_core import describe, annotate, add_swagger
import json


@describe(paths="/foo/{multiplier}")
@annotate({"multiplier": int, "return": int})
def get(self, multiplier):
    return 2 * multiplier


@describe(paths="/add")
@annotate({"left": int, "right": int, "return": int})
def add(self, left, right):
    return left + right


@describe(paths="/add", methods="POST")
@annotate({"left": int, "right": int, "return": int})
def subtract(self, left, right):
    return left - right


@describe(paths="/exception", methods="GET")
def exception(self):
    return raise_exception()


def raise_exception():
    raise Exception("OFOOONTHE")


app = route_set = RouteSet()
route_set.add(get)
route_set.add(add)
route_set.add(subtract)
route_set.add(exception)
app = tornado.web.Application(route_set.generate_url_specs())
add_swagger(app, "/swagger.json", "/swagger")

if __name__ == "__main__":
    app.listen(8888)
    tornado.ioloop.IOLoop.current().start()



from transmute_core.swagger import (
    generate_swagger_html,
    get_swagger_static_root,
    SwaggerSpec,
)
from transmute_core import default_context
import tornado
import tornado.web

USE_ROUTER = tornado.version_info >= (4, 5)
METHODS = ["get", "post", "delete", "put"]
STATIC_ROOT = "/_swagger/static"


def add_swagger(app, json_route, html_route, context=default_context):
    app.add_handlers(
        ".*",
        [
            (json_route, generate_swagger_json_handler(app, context)),
        ]
        + add_swagger_api_route(html_route, json_route),
    )


def generate_swagger_json_handler(app, context, **kwargs):
    swagger_json = _generate_swagger_json(app, context, **kwargs)

    class SwaggerSpecHandler(tornado.web.RequestHandler):
        def get(self):
            self.write(swagger_json)
            self.finish()

    return SwaggerSpecHandler


def _get_handlers(app):
    if USE_ROUTER:
        for rule in app.wildcard_router.rules:
            yield rule.target
    else:
        for domain, specs in app.handlers:
            for s in specs:
                yield s.handler_class


def _generate_swagger_json(app, context, **kwargs):
    spec = SwaggerSpec()
    for handler in _get_handlers(app):
        for m in METHODS:
            method = getattr(handler, m)
            if hasattr(method, "transmute_func"):
                spec.add_func(method.transmute_func, context)
    return spec.swagger_definition(**kwargs)


def add_swagger_api_route(target_route, swagger_json_route):
    static_root = get_swagger_static_root()
    swagger_body = generate_swagger_html(STATIC_ROOT, swagger_json_route)

    class SwaggerBodyHandler(tornado.web.RequestHandler):
        def get(self):
            self.write(swagger_body)
            self.finish()

    return [
        (target_route, SwaggerBodyHandler),
        (STATIC_ROOT + "/(.*)", tornado.web.StaticFileHandler, {"path": static_root}),
    ]



"""
An example integration with flask.
"""
import json
import sys
import transmute_core
import attr
from transmute_core import (
    describe,
    annotate,
    default_context,
    generate_swagger_html,
    get_swagger_static_root,
    ParamExtractor,
    SwaggerSpec,
    TransmuteFunction,
    NoArgument,
)
from flask import Blueprint, Flask, Response, request
from schematics.models import Model
from schematics.types import StringType
from functools import wraps

SWAGGER_ATTR_NAME = "_tranmute_swagger"
STATIC_PATH = "/_swagger/static"


def transmute_route(app, fn, context=default_context):
    """
    this is the main interface to transmute. It will handle
    adding converting the python function into the a flask-compatible route,
    and adding it to the application.
    """
    transmute_func = TransmuteFunction(fn)
    routes, handler = create_routes_and_handler(transmute_func, context)
    for r in routes:
        """
        the route being attached is a great place to start building up a
        swagger spec.  the SwaggerSpec object handles creating the
        swagger spec from transmute routes for you.

        almost all web frameworks provide some app-specific context
        that one can add values to. It's recommended to attach
        and retrieve the swagger spec from there.
        """
        if not hasattr(app, SWAGGER_ATTR_NAME):
            setattr(app, SWAGGER_ATTR_NAME, SwaggerSpec())
        swagger_obj = getattr(app, SWAGGER_ATTR_NAME)
        swagger_obj.add_func(transmute_func, context)
        app.route(r, methods=transmute_func.methods)(handler)


def create_routes_and_handler(transmute_func, context):
    """
    return back a handler that is the api generated
    from the transmute_func, and a list of routes
    it should be mounted to.
    """

    @wraps(transmute_func.raw_func)
    def handler():
        exc, result = None, None
        try:
            args, kwargs = ParamExtractorFlask().extract_params(
                context, transmute_func, request.content_type
            )
            result = transmute_func(*args, **kwargs)
        except Exception as e:
            exc = e
            """
            attaching the traceack is done for you in Python 3, but
            in Python 2 the __traceback__ must be
            attached to the object manually.
            """
            exc.__traceback__ = sys.exc_info()[2]
        """
        transmute_func.process_result handles converting
        the response from the function into the response body,
        the status code that should be returned, and the
        response content-type.
        """
        response = transmute_func.process_result(
            context, result, exc, request.content_type
        )
        return Response(
            response["body"],
            status=response["code"],
            mimetype=response["content-type"],
            headers=response["headers"],
        )

    return (_convert_paths_to_flask(transmute_func.paths), handler)


def _convert_paths_to_flask(transmute_paths):
    """
    convert transmute-core's path syntax (which uses {var} as the
    variable wildcard) into flask's <var>.
    """
    paths = []
    for p in transmute_paths:
        paths.append(p.replace("{", "<").replace("}", ">"))
    return paths


class ParamExtractorFlask(ParamExtractor):
    """
    The code that converts http parameters into function signature
    arguments is complex, so the abstract class ParamExtractor is
    provided as a convenience.

    override the methods to complete the class.
    """

    def __init__(self, *args, **kwargs):
        """
        in the case of flask, this is blank. But it's common
        to pass request-specific variables in the ParamExtractor,
        to be used in the methods.
        """
        super(ParamExtractorFlask, self).__init__(*args, **kwargs)

    def _get_framework_args(self):
        """
        this method should return back a dictionary of the values that
        are normally passed into the handler (e.g. the "request" object
        in aiohttp).

        in the case of flask, this is blank.
        """
        return {}

    @property
    def body(self):
        return request.get_data()

    @staticmethod
    def _query_argument(key, is_list):
        if key not in request.args:
            return NoArgument
        if is_list:
            return request.args.getlist(key)
        else:
            return request.args[key]

    @staticmethod
    def _header_argument(key):
        return request.headers.get(key, NoArgument)

    @staticmethod
    def _path_argument(key):
        return request.match_info.get(key, NoArgument)


def add_swagger(app, json_route, html_route, **kwargs):
    """
    add a swagger html page, and a swagger.json generated
    from the routes added to the app.
    """
    spec = getattr(app, SWAGGER_ATTR_NAME)
    if spec:
        spec = spec.swagger_definition(**kwargs)
    else:
        spec = {}
    encoded_spec = json.dumps(spec).encode("UTF-8")

    @app.route(json_route)
    def swagger():
        return Response(
            encoded_spec,
            # we allow CORS, so this can be requested at swagger.io
            headers={"Access-Control-Allow-Origin": "*"},
            content_type="application/json",
        )

    # add the statics
    static_root = get_swagger_static_root()
    swagger_body = generate_swagger_html(STATIC_PATH, json_route).encode("utf-8")

    @app.route(html_route)
    def swagger_ui():
        return Response(swagger_body, content_type="text/html")

    # the blueprint work is the easiest way to integrate a static
    # directory into flask.
    blueprint = Blueprint(
        "swagger", __name__, static_url_path=STATIC_PATH, static_folder=static_root
    )
    app.register_blueprint(blueprint)


# example usage.


@describe(
    paths="/api/v1/multiply/{document_id}",
    header_parameters=["header"],
    body_parameters="foo",
)
@annotate(
    {
        "left": int,
        "right": int,
        "header": int,
        "foo": str,
        "return": int,
        "document_id": str,
    }
)
def multiply(left, right, foo, document_id, header=0):
    return left * right


@describe(paths="/api/v1/multiply_body", body_parameters="body")
@annotate({"body": int})
def multiply_body(body):
    return left * right


@describe(paths="/api/v1/test")
@annotate({"vals": [int], "return": [int]})
def foo(vals):
    return vals


class SchematicsBody(Model):
    name = StringType(max_length=5)


@describe(
    paths="/api/v1/schematics", methods=["POST"], tags=["foo"], body_parameters="body"
)
@annotate({"body": SchematicsBody})
def schematics_example(body):
    return None


@describe(
    paths="/api/v1/header",
    response_types={
        200: {
            "type": str,
            "description": "success",
            "headers": {
                "location": {"description": "url to the location", "type": str}
            },
        },
    },
)
def header():
    return transmute_core.Response("foo", headers={"x-nothing": "value"})


@attr.s
class AttrsExample(object):
    foo = attr.ib(type=str)


@describe(paths="/api/v1/attrs")
@annotate({"return": AttrsExample})
def attrs():
    return AttrsExample(foo="bar")


app = Flask(__name__)
app = Flask(__name__)
transmute_route(app, attrs)
transmute_route(app, multiply)
transmute_route(app, multiply_body)
transmute_route(app, schematics_example)
transmute_route(app, foo)
transmute_route(app, header)
add_swagger(app, "/api/swagger.json", "/api/")

if __name__ == "__main__":
    app.run(debug=True)



from .handler import convert_to_handler
from .url import url_spec
from transmute_core import default_context
import tornado.web


class RouteSet(object):
    def __init__(self, context=default_context):
        self._handler_by_path = {}
        self._context = context

    def add(self, func):
        handler = convert_to_handler(self._context)(func)
        transmute_func = handler.transmute_func
        for p in transmute_func.paths:
            if p not in self._handler_by_path:

                class TransmuteRequestHandler(tornado.web.RequestHandler):
                    pass

                self._handler_by_path[p] = TransmuteRequestHandler
            request_class = self._handler_by_path[p]
            for method in transmute_func.methods:
                setattr(request_class, method.lower(), handler)

    def generate_url_specs(self):
        routes = []
        for path, handler in self._handler_by_path.items():
            routes.append(url_spec(path, handler))
        return routes



import json


def test_happy_path(test_app):
    r = test_app.get("/multiply?left=3&right=3")
    assert json.loads(r.data.decode()) == 9


def test_headers(test_app):
    r = test_app.get("/api/v1/header")
    assert r.headers["x-nothing"] == "value"


def test_complex(test_app):
    r = test_app.post(
        "/complex/3",
        data=json.dumps({"body": "1"}),
        headers={"header": "2", "content-type": "application/json"},
    )
    assert json.loads(r.data.decode()) == "1:2:3"


def test_api_exception(test_app):
    r = test_app.get("/exception")
    assert r.status_code == 400
    resp = json.loads(r.data.decode())
    assert resp["code"] == 400
    assert resp["success"] is False


def test_swagger(test_app):
    r = test_app.get("/swagger.json")
    swagger = json.loads(r.data.decode())
    assert swagger["info"] == {"version": "1.0", "title": "example"}
    assert "/multiply" in swagger["paths"]
    assert "/exception" in swagger["paths"]
    # test blueprint is documented as well
    assert "/blueprint/foo" in swagger["paths"]


def test_swagger_html(test_app):
    r = test_app.get("/api/")
    assert "/swagger.json" in r.data.decode()
    assert r.status_code == 200



from functools import wraps
from aiohttp.web import HTTPException
from transmute_core.exceptions import APIException, NoSerializerFound
from transmute_core.function.signature import NoDefault
from transmute_core import ParamExtractor, NoArgument
from aiohttp import web


def create_handler(transmute_func, context):
    @wraps(transmute_func.raw_func)
    async def handler(request):
        exc, result = None, None
        try:
            args, kwargs = await extract_params(request, context, transmute_func)
            result = await transmute_func.raw_func(*args, **kwargs)
        except HTTPException as hpe:
            code = hpe.status_code or 400
            exc = APIException(code=code, message=str(hpe))
        except Exception as e:
            exc = e
        if isinstance(result, web.Response):
            return result
        else:
            response = transmute_func.process_result(
                context, result, exc, request.content_type
            )
            return web.Response(
                body=response["body"],
                status=response["code"],
                content_type=response["content-type"],
                headers=response["headers"],
            )

    handler.transmute_func = transmute_func
    return handler


async def extract_params(request, context, transmute_func):
    body = await request.content.read()
    content_type = request.content_type
    extractor = ParamExtractorAIOHTTP(request, body)
    return extractor.extract_params(context, transmute_func, content_type)


class ParamExtractorAIOHTTP(ParamExtractor):
    def __init__(self, request, body):
        self._request = request
        self._body = body

    def _get_framework_args(self):
        return {"request": self._request}

    @property
    def body(self):
        return self._body

    def _query_argument(self, key, is_list):
        if key not in self._request.query:
            return NoArgument
        if is_list:
            return self._request.query.getall(key)
        else:
            return self._request.query[key]

    def _header_argument(self, key):
        return self._request.headers.get(key, NoArgument)

    def _path_argument(self, key):
        return self._request.match_info.get(key, NoArgument)



import json
from aiohttp import web

from transmute_core.swagger import (
    generate_swagger_html,
    get_swagger_static_root,
    SwaggerSpec,
)

STATIC_ROOT = "/_swagger/static"
APP_KEY = "_aiohttp_transmute_swagger"


def get_swagger_spec(app):
    if APP_KEY not in app:
        app[APP_KEY] = SwaggerSpec()
    return app[APP_KEY]


def add_swagger(app, json_route, html_route):
    """
    a convenience method for both adding a swagger.json route,
    as well as adding a page showing the html documentation
    """
    app.router.add_route("GET", json_route, create_swagger_json_handler(app))
    add_swagger_api_route(app, html_route, json_route)


def add_swagger_api_route(app, target_route, swagger_json_route):
    """
    mount a swagger statics page.

    app: the aiohttp app object
    target_route: the path to mount the statics page.
    swagger_json_route: the path where the swagger json definitions is
                        expected to be.
    """
    static_root = get_swagger_static_root()
    swagger_body = generate_swagger_html(STATIC_ROOT, swagger_json_route).encode(
        "utf-8"
    )

    async def swagger_ui(request):
        return web.Response(body=swagger_body, content_type="text/html")

    app.router.add_route("GET", target_route, swagger_ui)
    app.router.add_static(STATIC_ROOT, static_root)


def create_swagger_json_handler(app, **kwargs):
    """
    Create a handler that returns the swagger definition
    for an application.

    This method assumes the application is using the
    TransmuteUrlDispatcher as the router.
    """

    spec = get_swagger_spec(app).swagger_definition(**kwargs)
    encoded_spec = json.dumps(spec).encode("UTF-8")

    async def swagger(request):
        return web.Response(
            # we allow CORS, so this can be requested at swagger.io
            headers={"Access-Control-Allow-Origin": "*"},
            body=encoded_spec,
            content_type="application/json",
        )

    return swagger



import tornado.testing
import tornado.web
from transmute_core import describe, annotate
from transmute_core.frameworks.tornado import RouteSet, add_swagger
import json

# frameworks.tornado also provides a RouteSet, which can output tornado-compatible url objects.
# this provides a more flask-like approach, and handles combining routes which have the same path


@describe(paths="/foo/{multiplier}")
@annotate({"multiplier": int, "return": int})
def get(self, multiplier):
    return 2 * multiplier


@describe(paths="/add")
@annotate({"left": int, "right": int, "return": int})
def add(self, left, right):
    return left + right


@describe(paths="/add", methods="POST")
@annotate({"left": int, "right": int, "return": int})
def subtract(self, left, right):
    return left - right


class TestApp(tornado.testing.AsyncHTTPTestCase):
    def get_app(self):
        route_set = RouteSet()
        route_set.add(get)
        route_set.add(add)
        route_set.add(subtract)
        app = tornado.web.Application(route_set.generate_url_specs())
        add_swagger(app, "/swagger.json", "/swagger")
        return app

    def test_foo(self):
        resp = self.fetch("/foo/1")
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert resp_json == 2

    def test_queryparam(self):
        resp = self.fetch("/add?left=1&right=10")
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert resp_json == 11

    def test_missing_arg(self):
        resp = self.fetch("/add?left=1")
        assert resp.code == 400

    def test_body(self):
        resp = self.fetch(
            "/add",
            method="POST",
            body=json.dumps({"left": 10, "right": 2}),
            headers={"Content-Type": "application/json"},
        )
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert resp_json == 8

    def test_swagger_json(self):
        resp = self.fetch("/swagger.json")
        assert resp.code == 200
        resp_json = json.loads(resp.body.decode("UTF-8"))
        assert "get" in resp_json["paths"]["/foo/{multiplier}"]
        assert resp_json["swagger"] == "2.0"
        assert resp_json["info"] == {"title": "example", "version": "1.0"}

    def test_swagger_body(self):
        resp = self.fetch("/swagger")
        assert resp.code == 200
        body = resp.body.decode("UTF-8")
        assert "End Swagger" in body
